
#pragma once
#include <iostream>
using namespace std;

#include <stdio.h>

#include "MyException.h"

template <class T>
class DList
{
private:
   struct DListNode
   {
      T value;
      DListNode *next, *prev;
      DListNode() { next = prev = 0; }
      void LinkNext(DListNode *p) { next = p; if (p) p->prev = this; }
      void LinkPrev(DListNode *p) { prev = p; if (p) p->next = this; }
   };

   DListNode *currentNode;
   DListNode *first, *last;
   int border;   //      2-, 1-
   int size;

public:
   // , 
   DList() { currentNode = first = last = 0; size = 0; border = 3; }
   ~DList();

   // 
   void InsAfter(const T &x);
   void InsBefore(const T &x);
   void PushFirst(const T &x);
   void PushLast(const T &x);

   // 
   void DelAfter();
   void DelBefore();
   void DelCurForAfter();
   void DelCurForBefore();

   // 
   T & Value()   { return currentNode->value; }
   const T & Value() const { return currentNode->value; }

   //   
   void ToNext();
   void ToPrev();
   void ToFirst();
   void ToLast();

   //  
   bool AtBegin() const { return border & 2; }   //   " "
   bool AtEnd() const { return border & 1; }
   int  Size() const { return size; }  

   template <class P>
   friend ostream & operator<<(ostream & os, const DList<P> & list);

};

 template <class T>
 ostream & operator<<(ostream & os, const DList<T> & list)
 {
    typename DList<T>::DListNode *p;
    cout << "[" << list.size << "," << list.border << "] ";
    for (p  = list.first; p; p = p->next) {
       cout << " " << p->value ;
    }
    cout << endl;
    return os;
 }  


 template <class T>
 DList<T>::~DList()
 {
    DListNode *p = first;
    while (first) {
        p = first;
        first = first->next;
        delete p;
    }
 }

                                                    
 template <class T>
 void DList<T>::InsAfter(const T &x)
 {
  try {
     if (border == 1) return;
     DListNode *p = new DListNode;
     p->value = x;
     switch(border) 
     {
      case 0:
         p->LinkNext(currentNode->next);
         p->LinkPrev(currentNode);
         if (currentNode == last) { last = p; }
         break;
      case 2:
         p->LinkNext(first);
         first = p;
         break;
      case 3:
         first = last = p;
         border = 2;
         break;
     }
     ++size;
  } catch (std::bad_alloc) {
     throw new MyException ("DList<T>::InsAfter: memory allocation error\n", EC_MEMORY);
  }                                                            
 } 

 template <class T>
 void DList<T>::DelAfter()
 {
     DListNode *p;
     switch(border) {
      case 0:
         p = currentNode->next;           
         if (p) {
             currentNode->LinkNext(p->next);
             delete p;
             --size;
             if (currentNode->next == 0) {
                last = currentNode;
             }
         }
      case 2:
         p = first;           
         first = first->next; 
         delete p;
         --size;
         if (first == 0) {
            last = 0;
            border = 3;
         }
         break;
      case 1:
      case 3:
         return;
     }
 } 


 template <class T>
 void DList<T>::ToFirst()
 {
    currentNode = first;
    if (first) border = 0;
 } 

 template <class T>
 void DList<T>::ToNext()
 {
    if (size == 0 || border & 1) return;
    if (border == 2) {
       currentNode = first;
       border = 0;
    } else {
       currentNode = currentNode->next;
       if (currentNode == 0) border = 1; 
    }
 } 



